    {"scandir",         (PyCFunction)posix_scandir,
                        METH_VARARGS | METH_KEYWORDS,
                        posix_scandir__doc__},
    {"listdir2",        (PyCFunction)posix_listdir2,
                        METH_VARARGS | METH_KEYWORDS,
                        NULL},
